#include "main.h"

void 
write_ban_file(void)
{
  FILE *fp;
  struct ban_item *b;
 
  if( (fp = fopen(gs.banfile, "w")) == NULL ) {
    print_to_log("Unable to open ban file '%s' for writing\n", gs.banfile);
    return;
  }
 
  b = gs.banip_list;
  while( b ) {
    fprintf(fp, "%-11s %s %s %s\n", "ban_ip", b->ipaddr, b->banner, b->bannee);
    b = b->next;
  }

  b = gs.banexclude_list;
  while( b ) {
    fprintf(fp, "%-11s %s %s %s\n", "ban_exclude", b->ipaddr, b->banner, b->bannee);
    b = b->next;
  }

  b = gs.banname_list;
  while( b ) {
    fprintf(fp, "%-11s %s %s %s\n", "ban_name", b->regex, b->banner, b->bannee);
    b = b->next;
  }

  b = gs.bancolor_list;
  while( b ) {
    fprintf(fp, "%-11s %-2d %-2d\n", "ban_color", b->shirt, b->pants);
    b = b->next;
  }

  fclose(fp);
}

void 
read_ban_file(void)
{
    FILE *fp;
    char line[400], ipaddr[100], regex[100], banner[100], bannee[100];
    int shirt, pants;
 
    if( (fp = fopen(gs.banfile, "r")) == NULL ) {
      print_to_log("Couldn't open ban file '%s' for reading\n", gs.banfile);
      return;
    }
 
    while( fgets(line, 400, fp) != NULL ) {

      regex[0] = ipaddr[0] = banner[0] = bannee[0] = '\0';

      if( sscanf(line, "ban_ip %s %s %[^\n]", ipaddr, banner, bannee) ) {
         add_ban_ip_item(ipaddr, banner, bannee);
      } else if( sscanf(line, "ban_exclude %s %s %[^\n]",  ipaddr, banner, bannee) ) {
         add_ban_exclude_item(ipaddr, banner, bannee);
      } else if( sscanf(line, "ban_name %s %s %[^\n]",  regex, banner, bannee) ) {
         add_ban_name_item(regex, banner, bannee);
      } else if( sscanf(line, "ban_color %d %d", &shirt, &pants) == 2 ) {
         add_ban_color_item(shirt, pants);
      } else {
         print_to_log("Unknown or bad command in ban file: %s", line);
      }
    }
}

void 
destroy_ban_lists(void)
{
  struct ban_item *b;

  while( gs.banip_list ) {
    b = gs.banip_list;
    gs.banip_list = gs.banip_list->next;
    destroy_ban_item(b);
  }
  while( gs.banexclude_list ) {
    b = gs.banexclude_list;
    gs.banexclude_list = gs.banexclude_list->next;
    destroy_ban_item(b);
  }
  while( gs.banname_list ) {
    b = gs.banname_list;
    gs.banname_list = gs.banname_list->next;
    destroy_ban_item(b);
  }
  while( gs.bancolor_list ) {
    b = gs.bancolor_list;
    gs.bancolor_list = gs.bancolor_list->next;
    destroy_ban_item(b);
  }
}

void 
destroy_ban_item(struct ban_item *b)
{
  if( b->ipaddr ) free(b->ipaddr);
  if( b->banner ) free(b->banner);
  if( b->bannee ) free(b->bannee);
  if( b->regex  ) {
    free(b->regex);
    regfree(&b->regcmp);
  }
  free(b);
}

struct ban_item *
make_ban_item(void)
{
  struct ban_item *new;

  if ((new = (struct ban_item *)malloc(sizeof(struct ban_item))) == NULL) {
      print_to_log("make_ban_item: malloc error");
      shutdown_all();
  }
  new->regex  = NULL;
  new->ipaddr = "";
  new->shirt  = 0;
  new->pants  = 0;
  new->bannee = "";
  new->banner = "";
  new->next   = NULL;

  return new;
}

int 
map_ip_to_regex(char *ipaddr, char *regex)
{
   char *cp = ipaddr;
   int periods = 0;
 
   strcpy(regex, "^");
   while( *cp != '\0' && *cp != '\n' ) {
     if( isdigit((int)*cp) ) {
       strncat(regex, cp, 1);
       cp++;
     } else if( *cp == '.' ) {
       strcat(regex, "[.]");
       cp++;
       periods++;
     } else if( *cp == '*' ) {
       strcat(regex, "[0-9]{1,3}");
       cp++;
     } else {
       return 0;
     }
   }
   strcat(regex, "$");

   if( strlen(regex) < 15 || periods != 3) {
     return 0;
   } else {
     return 1;
   }
}

void 
add_ban_ip_item(char *ipaddr, char *banner, char *bannee)
{
  struct ban_item *new;
  char regex[100];

  if( ipaddr ) {
    if( !map_ip_to_regex(ipaddr, regex) ) {
      print_to_log("Improper ipaddr in ban file: %s\n", ipaddr);
      return;
    }
    remove_banip_item_by_regex(regex);
    new         = make_ban_item();
    new->regex  = strdup(regex);
    new->ipaddr = strdup(ipaddr);
    if( bannee ) {
      new->bannee = strdup(bannee);
    }
    if( banner ) {
      new->banner = strdup(banner);
    }
    if( regcomp(&new->regcmp, new->regex, REG_EXTENDED) ) {
      send_to_clients("commandprint Improper ban ipaddr: %s\n", new->ipaddr);
      print_to_log("Improper ban ipaddr: %s\n", new->ipaddr);
      destroy_ban_item(new);
      return;
    }

    new->next     = gs.banip_list;
    gs.banip_list = new;
    send_to_clients("commandprint Ipaddr %s added to ban list\n", new->ipaddr);
    print_to_log("Ipaddr %s added to ban list\n", new->ipaddr);
  } else {
    print_to_log("Null ipaddr or regex passed for ip ban item\n");
  }
}

void 
add_ban_exclude_item(char *ipaddr, char *banner, char *bannee)
{
  struct ban_item *new;
  char regex[100];

  if( ipaddr ) {
    if( !map_ip_to_regex(ipaddr, regex) ) {
      print_to_log("Improper ipaddr in ban file: %s\n", ipaddr);
      return;
    }
    remove_banexclude_item_by_regex(regex);
    new         = make_ban_item();
    new->regex  = strdup(regex);
    new->ipaddr = strdup(ipaddr);
    if( bannee ) {
      new->bannee = strdup(bannee);
    }
    if( banner ) {
      new->banner = strdup(banner);
    }
    if( regcomp(&new->regcmp, new->regex, REG_EXTENDED) ) {
      send_to_clients("commandprint Improper ban ipaddr: %s\n", new->ipaddr);
      print_to_log("Improper ban ipaddr: %s\n", new->ipaddr);
      destroy_ban_item(new);
      return;
    }

    new->next          = gs.banexclude_list;
    gs.banexclude_list = new;
    send_to_clients("commandprint Ipaddr %s is now excluded from IP banning\n", new->ipaddr);
    print_to_log("Ipaddr %s is now excluded from IP banning\n", new->ipaddr);
  } else {
    print_to_log("Null ipaddr or regex passed for ip ban item\n");
  }
}

void 
add_ban_name_item(char *regex, char *banner, char *bannee)
{
  struct ban_item *new;
  unsigned char *tmp;

  if( regex ) {
    remove_banname_item_by_regex(regex);
    new = make_ban_item();
    new->regex  = strdup(regex);
    tmp = (unsigned char *)strdup(regex);
    map_control_chars(tmp, regex, strlen(tmp));
    if( bannee ) {
      new->bannee = strdup(bannee);
    }
    if( banner ) {
      new->banner = strdup(banner);
    }
    if( regcomp(&new->regcmp, tmp, REG_EXTENDED | REG_ICASE) ) {
      send_to_clients("Improper ban name: %s\n", new->regex);
      print_to_log("Improper ban name: %s\n", new->regex);
      destroy_ban_item(new);
    }

    new->next   = gs.banname_list;
    gs.banname_list = new;
    send_to_clients("commandprint Name regex %s added to ban list\n", new->regex);
    print_to_log("Name regex %s added to ban list\n", new->regex);
    free(tmp);
  } else {
    print_to_log("Null regex passed for name ban item\n");
  }
}

void 
add_ban_color_item(int shirt, int pants)
{
  struct ban_item *new;

  if( shirt >=0 && shirt <= 13 && pants >= 0 && pants <= 13  ) {
    remove_bancolor_item_by_color(shirt, pants);
    new              = make_ban_item();
    new->shirt       = shirt;
    new->pants       = pants;
    new->next        = gs.bancolor_list;
    gs.bancolor_list = new;
    send_to_clients("commandprint Colors '%d %d' added to ban list\n", 
           new->shirt, new->pants);
    print_to_log("Colors '%d %d' added to ban list\n", 
           new->shirt, new->pants);
  } else {
    print_to_log("Bad arguments to ban_color command: %d %d\n", shirt, pants);
  }
}

void 
remove_bancolor_item_by_color(int shirt, int pants)
{
  struct ban_item *b, *tmp;
 
  if( gs.bancolor_list != NULL) {
    if( gs.bancolor_list->shirt == shirt && gs.bancolor_list->pants == pants ) {
       tmp = gs.bancolor_list;
       gs.bancolor_list = gs.bancolor_list->next;
       destroy_ban_item(tmp);
       return;
    } else {
      b = gs.bancolor_list;
      while( b != NULL && b->next != NULL ) {
        if( b->next->shirt == shirt && b->next->pants == pants ) {
          tmp = b->next;
          b->next = b->next->next;
          destroy_ban_item(tmp);
          return;
        }
        b = b->next;
      }
    }
  }
}

void 
remove_banip_item_by_regex(char *regex)
{
  struct ban_item *b, *tmp;
 
  if( regex != NULL && gs.banip_list != NULL) {
    if( strcmp(gs.banip_list->regex, regex) == 0 ) {
       tmp = gs.banip_list;
       gs.banip_list = gs.banip_list->next;
       destroy_ban_item(tmp);
       return;
    } else {
      b = gs.banip_list;
      while( b != NULL && b->next != NULL ) {
        if( strcmp(b->next->regex, regex) == 0 ) {
          tmp = b->next;
          b->next = b->next->next;
          destroy_ban_item(tmp);
          return;
        }
        b = b->next;
      }
    }
  }
}

void 
remove_banexclude_item_by_regex(char *regex)
{
  struct ban_item *b, *tmp;
 
  if( regex != NULL && gs.banexclude_list != NULL) {
    if( strcmp(gs.banexclude_list->regex, regex) == 0 ) {
       tmp = gs.banexclude_list;
       gs.banexclude_list = gs.banexclude_list->next;
       destroy_ban_item(tmp);
       return;
    } else {
      b = gs.banexclude_list;
      while( b != NULL && b->next != NULL ) {
        if( strcmp(b->next->regex, regex) == 0 ) {
          tmp = b->next;
          b->next = b->next->next;
          destroy_ban_item(tmp);
          return;
        }
        b = b->next;
      }
    }
  }
}

void 
remove_banname_item_by_regex(char *regex)
{
  struct ban_item *b, *tmp;
 
  if( regex != NULL && gs.banname_list != NULL) {
    if( strcmp(gs.banname_list->regex, regex) == 0 ) {
       tmp = gs.banname_list;
       gs.banname_list = gs.banname_list->next;
       destroy_ban_item(tmp);
       return;
    } else {
      b = gs.banname_list;
      while( b != NULL && b->next != NULL ) {
        if( strcmp(b->next->regex, regex) == 0 ) {
          tmp = b->next;
          b->next = b->next->next;
          destroy_ban_item(tmp);
          return;
        }
        b = b->next;
      }
    }
  }
}

void 
remove_ban_item_by_num(int num)
{
  struct ban_item *b, *tmp = NULL;
  int i = 1;

  if( num > 0 && gs.banip_list != NULL) {
    if( num == i++ ) {
       tmp = gs.banip_list;
       gs.banip_list = gs.banip_list->next;
    } else {
      b = gs.banip_list;
      while( b != NULL && b->next != NULL ) {
        if( i == num ) {
          tmp = b->next;
          b->next = b->next->next;
          break;
        }
        b = b->next;
        i++;
      }
    }
    if( tmp ) {
      send_to_clients("commandprint Ipaddr %s removed from ban list\n", tmp->ipaddr);
      print_to_log("Ipaddr %s removed from ban list\n", tmp->ipaddr);
      destroy_ban_item(tmp);
      return;
    }
  }

  if( num > 0 && gs.banexclude_list != NULL) {
    if( num == i++ ) {
       tmp = gs.banexclude_list;
       gs.banexclude_list = gs.banexclude_list->next;
    } else {
      b = gs.banexclude_list;
      while( b != NULL && b->next != NULL ) {
        if( i == num ) {
          tmp = b->next;
          b->next = b->next->next;
          break;
        }
        b = b->next;
        i++;
      }
    }
    if( tmp ) {
      send_to_clients("commandprint Ipaddr %s removed from excluded ban list\n", 
                tmp->ipaddr);
      print_to_log("Ipaddr %s removed from excluded ban list\n", tmp->ipaddr);
      destroy_ban_item(tmp);
      return;
    }
  }

  if( num > 0 && gs.banname_list != NULL) {
    if( num == i++ ) {
       tmp = gs.banname_list;
       gs.banname_list = gs.banname_list->next;
    } else {
      b = gs.banname_list;
      while( b != NULL && b->next != NULL ) {
        if( i == num ) {
          tmp = b->next;
          b->next = b->next->next;
          break;
        }
        b = b->next;
        i++;
      }
    }
    if( tmp ) {
      send_to_clients("commandprint Name regex '%s' removed from ban list\n", tmp->regex);
      print_to_log("Name regex '%s' removed from ban list\n", tmp->regex);
      destroy_ban_item(tmp);
      return;
    }
  }

  if( num > 0 && gs.bancolor_list != NULL) {
    if( num == i++ ) {
       tmp = gs.bancolor_list;
       gs.bancolor_list = gs.bancolor_list->next;
    } else {
      b = gs.bancolor_list;
      while( b != NULL && b->next != NULL ) {
        if( i == num ) {
          tmp = b->next;
          b->next = b->next->next;
          break;
        }
        b = b->next;
        i++;
      }
    }
    if( tmp ) {
      send_to_clients("commandprint Colors '%d %d' removed from ban list\n", 
             tmp->shirt, tmp->pants);
      print_to_log("Colors '%d %d' removed from ban list\n", tmp->shirt, tmp->pants);
      destroy_ban_item(tmp);
      return;
    }
  }
  send_to_clients("commandprint No names, ips, or colors at number '%d' in ban lists\n", num);
  print_to_log("No names, ips, or colors at number '%d' in ban lists\n", num);
}

void 
say_ban_list(void)
{
  struct ban_item *b;
  int i;

  if( gs.banip_list || gs.banexclude_list || gs.banname_list || gs.bancolor_list ) {
    i = 1;
    if( gs.banip_list ) {
      send_console_command("say Banned Ips:");
      b = gs.banip_list;
      while( b ) {
        send_console_command("say %2d %-15s", i, b->ipaddr);
        b = b->next;
        i++;
      }
    }
    if( gs.banexclude_list ) {
      send_console_command("say Excluded Ips:");
      b = gs.banexclude_list;
      while( b ) {
        send_console_command("say %2d %-15s", i, b->ipaddr);
        b = b->next;
        i++;
      }
    }
    if( gs.banname_list ) {
      send_console_command("say Banned Names:");
      b = gs.banname_list;
      while( b ) {
        send_console_command("say %2d %-15s", i, b->regex);
        b = b->next;
        i++;
      }
    }
    if( gs.bancolor_list ) {
      send_console_command("say Banned Colors:");
      b = gs.bancolor_list;
      while( b ) {
        send_console_command("say %2d %2d %2d", i, b->shirt, b->pants);
        b = b->next;
        i++;
      }
    }
  } else {
    send_console_command("say No names, IPs or colors are currently banned or excluded");
  }
}

void 
send_ban_list(int index)
{
  struct ban_item *b;
  int i;

  if( gs.banip_list || gs.banname_list || gs.bancolor_list ) {
    i = 1;
    if( gs.banip_list ) {
      send_to_client(index, "commandprint    IP              Banner                Bannee\n");
      b = gs.banip_list;
      while( b ) {
        send_to_client(index, "commandprint %2d %-15s %-21s %s\n", 
                        i, b->ipaddr, b->banner, b->bannee);
        b = b->next;
        i++;
      }
    }
    if( gs.banexclude_list ) {
      send_to_client(index, "commandprint    Excluded IP     Excluder            Excludee\n");
      b = gs.banexclude_list;
      while( b ) {
        send_to_client(index, "commandprint %2d %-15s %-21s %s\n", 
                        i, b->ipaddr, b->banner, b->bannee);
        b = b->next;
        i++;
      }
    }
    if( gs.banname_list ) {
      send_to_client(index, "commandprint    Name            Banner                Bannee\n");
      b = gs.banname_list;
      while( b ) {
        send_to_client(index, "commandprint %2d %-15s %-21s %s\n", 
                        i, b->regex, b->banner, b->bannee);
        b = b->next;
        i++;
      }
    }
    if( gs.bancolor_list ) {
      send_to_client(index, "commandprint    Shirt Pants\n");
      b = gs.bancolor_list;
      while( b ) {
        send_to_client(index, "commandprint %2d %2d   %2d\n", i, b->shirt, b->pants);
        b = b->next;
        i++;
      }
    }
  } else {
    send_to_client(index, "commandprint No names, IPs or colors are currently banned\n");
  }
}

int
ipaddr_is_excluded(char *ipaddr)
{
  struct ban_item *b;

  b = gs.banexclude_list;
  while( b ) {
    if( !regexec(&b->regcmp, ipaddr, 0, 0, REG_NOSUB) ) {
      return TRUE;
    }
    b = b->next;
  }
  return FALSE;
}

/* We can't send out an informative message to the victim of the IP ban because
 * we ban them before they really enter the game. Since they haven't told the 
 * server their name yet, we can't 'tell' them anything. Oh well, since their
 * IP is banned, they can't play just by changing their name or colors anyway. :)
 */
void
ban_check(void)
{
  int i;
  struct ban_item *b;

  for(i=0; i<MAX_PLAYERS; i++) {
    if( gs.players[i].valid && ((i+1) != gs.player_number) && gs.player_number != 0) {

      b = gs.banip_list;
      while( b ) {
        if( !regexec(&b->regcmp, gs.players[i].ipaddr, 0, 0, REG_NOSUB) &&
            !ipaddr_is_excluded(gs.players[i].ipaddr) )
        {
          send_to_clients("print Kicking banned IP: %s - %s, Banned by %s\n",
                  gs.players[i].ipaddr, gs.players[i].name, b->banner);
          print_to_log("Kicking banned IP: %s - %s, Banned by %s\n",
                  gs.players[i].ipaddr, gs.players[i].name, b->banner);
          send_impulse(i + 1 + gs.ban_impulse);
        }
        b = b->next;
      }

      b = gs.banname_list;
      while( b ) {
        if( !regexec(&b->regcmp, gs.players[i].name, 0, 0, REG_NOSUB) ) {
          send_to_clients("print Kicking banned name: %s - %s, Banned by %s\n",
                  b->regex, gs.players[i].name, b->banner);
          print_to_log("Kicking banned name: %s - %s, Banned by %s\n",
                  b->regex, gs.players[i].name, b->banner);
          send_console_command("tell \"%s\" You were banned for an illegal name!", 
              gs.players[i].real_name);
          send_impulse(i + 1 + gs.ban_impulse);
        }
        b = b->next;
      }

      b = gs.bancolor_list;
      while( b ) {
        if( gs.players[i].shirt == b->shirt && gs.players[i].pants == b->pants ) {
          send_to_clients("print Kicking banned colors: %d %d - %s\n",
                  b->shirt, b->pants, gs.players[i].name);
          print_to_log("Kicking banned colors: %d %d - %s\n",
                  b->shirt, b->pants, gs.players[i].name);
          send_console_command("tell \"%s\" You were banned for illegal colors!", 
              gs.players[i].real_name);
          send_impulse(i + 1 + gs.ban_impulse);
        }
        b = b->next;
      }

      if( gs.ban_white_guys >= 0 && gs.players[i].seconds >= gs.ban_white_guys &&
          gs.players[i].shirt == WHITE && gs.players[i].pants == WHITE )
      { 
        send_to_clients("print Kicking white guy - %s\n", gs.players[i].name);
        print_to_log("Kicking white guy - %s\n", gs.players[i].name);
        send_console_command("tell \"%s\" You were kicked for being a white guy!", 
            gs.players[i].real_name);
        send_console_command("tell \"%s\" Change your colors to anything else to get back."); 
        send_impulse(i + 1 + gs.ban_impulse);
      }
    }
  }
}

/*
 * We only want to send the banning impulse under certain conditions. If we
 * send the impulse at the wrong time, the server gets confused and prints
 * out an annoying message or doesn't give qsmack administrator priveleges.
 *
 * 1) We have send the init04 impulse (ban_lock is true)
 * 2) The ban impulse is set (from the config file)
 * 3) Our player number is set
 * 4) qsmack has been in the game for more than BAN_WAIT seconds
 * 5) The current level has been underway for more than BAN_WAIT seconds
 *
 * Condition 4) is necessary so the server has had time to grant qsmack admin
 * priveleges. If qsmack enters the game and immediately starts banning people
 * before it gets admin, the serving never grants qsmack admin and Eldie gets
 * pissed.
 *
 * Condition 5) isn't necessary because qsmack will already have admin, but it
 * prevents annoying messages that may be sent if qsmack tries to ban someone
 * during intermission.
 */
int 
ban_ok(void)
{
  if( gs.ban_lock == TRUE &&
      gs.ban_impulse >= 0 &&
      gs.player_number != 0 &&
      gs.players[gs.player_number-1].seconds > BAN_WAIT &&
      gs.time > BAN_WAIT
    )
  {
    return 1;
  } else {
/*     print_to_log("ban_ok failed - ban_lock:%d ban_impulse:%d mynumber:%d secs:%d time:%f\n", */
/*       gs.ban_lock, gs.ban_impulse, gs.player_number, gs.players[gs.player_number-1].seconds, */
/*       gs.time); */
    return 0;
  }
}

void 
map_control_chars(unsigned char *dst, unsigned char *src, int numchars)
{
   unsigned char tmp[4];
   unsigned int i, j, k;

   memset((unsigned char *)dst, 0, strlen(src));
   i = j = 0;
   while( i<strlen(src) && (int)j < numchars ) {
      if( src[i] == '\\' ) {
        i++;
        switch( src[i++] ) {
          case 'n':
            dst[j] = '\n';
            break;
          case 'r':
            dst[j] = '\r';
            break;
          case 't':
            dst[j] = '\t';
            break;
          case 'f':
            dst[j] = '\f';
            break;
          case 'b':
            dst[j] = '\b';
            break;
          case 'a':
            dst[j] = '\a';
            break;
          case 'x':
            for(k=0; k<2; k++) {
              if( isxdigit(src[i]) ) {
                tmp[k] = src[i++];
              } else {
                break;
              }
            }
            tmp[k] = '\0';
            dst[j] = (unsigned char)strtol(tmp, (char **)NULL, 16);
            break;
          case 'd':
            for(k=0; k<3; k++) {
              if( isdigit(src[i]) ) {
                tmp[k] = src[i++];
              } else {
                break;
              }
            }
            tmp[k] = '\0';
            dst[j] = (unsigned char)atoi(tmp);
            break;
          default:
            dst[j] = src[i-1];
        }
      } else {
        dst[j] = src[i++];
      }
      j++;
   }
   dst[j] = '\0';
}
