/*
 *  timer.c
 *
 *  Implements some stopwatch-like timer routines.
 *
 *  Kekoa Proudfoot
 *  12/12/96
 */

#include <stdio.h>
#include "timer.h"
#ifndef WIN32
#include <sys/time.h>
#else
#include <windows.h>
#include <sys/timeb.h>
#endif

#define FLOAT_TIME(tv)  (tv.tv_sec + tv.tv_usec * 1.0e-6)

#ifdef WIN32
int
gettimeofday(struct timeval *tp, void *tzp)
{
    struct timeb tb;
    ftime(&tb);
    tp->tv_sec  = tb.time;
    tp->tv_usec = tb.millitm * 1000;
    return 0;
}
#endif

double
reset_timer(qtimer_t *timer)
{
    timer->running      = 0;
    timer->elapsed_time = 0.0;

    return timer->elapsed_time;
}

double
start_timer(qtimer_t *timer)
{
    struct timeval now;

    if (timer->running)
	return read_timer(timer);

    gettimeofday(&now, NULL);

    timer->running    = 1;
    timer->start_time = FLOAT_TIME(now);

    return timer->elapsed_time;
}

double
stop_timer(qtimer_t *timer)
{
    struct timeval now;

    if (!timer->running)
	return timer->elapsed_time;

    gettimeofday(&now, NULL);
	
    timer->running = 0;
    timer->elapsed_time += FLOAT_TIME(now) - timer->start_time;

    return timer->elapsed_time;
}

double
read_timer(qtimer_t *timer)
{
    struct timeval now;

    if (!timer->running)
	return timer->elapsed_time;

    gettimeofday(&now, NULL);

    return FLOAT_TIME(now) - timer->start_time;
}

double
restart_timer(qtimer_t *timer)
{
    reset_timer(timer);
    return start_timer(timer);
}
